﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Cursed.geom;
using Cursed.gob.mob;
using Cursed.gob.mob.kob;

namespace Cursed
{
    public class LifeMap
    {
        const float F = 1f;
        const float oneOverB = .8f;

        int rows = Const.LifeMapRows; //number of rows
        int cols = Const.LifeMapCols; //number of cols

        CursedGame game;
        float[,] densities; //densities[i,j] is i'th row, j'th col

        Bounds bounds;

        float[] expLookup;
        const int ExpTableSize = 10000;

        int tick;

        float maxDensity;
        float minDensity;
        public static bool allGreen = false;
        public LifeMap(CursedGame game)
        {
            this.game = game;
            densities = new float[rows, cols];
            bounds = new Bounds(0, 0, 1, 1);

            expLookup = new float[ExpTableSize];
            for (int i = 0; i < ExpTableSize; i++)
            {
                expLookup[i] = (float)Math.Exp(-0.1f * i);
            }

            tick = 0;
        }

        float exp(float f)
        {
            int i = (int)f * -10;
            if (i < 0 || i >= ExpTableSize)
            {
                return (float)Math.Exp(f);
            }
            return expLookup[i];
        }

        float distance(Vector2 p1, Vector2 p2)
        {
            return F * exp(-(p1 - p2).LengthSquared() * oneOverB);
        }

        public float this[Vector2 loc]
        {
            get
            {
                float xt = (loc.X - bounds.MinX) / bounds.Width;
                float yt = (loc.Y - bounds.MinY) / bounds.Height;
                int row = (int)(yt * (rows - 1));
                int col = (int)(xt * (cols - 1));
                return this[row, col];
            }
        }

        public int Rows
        {
            get { return rows; }
        }

        public int Cols
        {
            get { return cols; }
        }

        public Point GetScreenPoint(int row, int col)
        {
            int x = (int)((float)col * bounds.Width / cols);
            int y = (int)((float)row * bounds.Height / rows);
            return new Point(x, y);
        }

        public float this[int row, int col]
        {
            get
            {
                if (row < 0 || row >= rows)
                    return 0;
                if (col < 0 || col >= cols)
                    return 0;
                float d = densities[row, col];
                if (d > 0)
                {
                    d /= Const.DruidEnergy;
                    d = Math.Min(1, d);
                }
                else
                {
                    d /= Const.DruidEnergy;
                    d = Math.Max(-1, d);
                }
                d *= 0.5f;
                d += 0.5f;
                return d;
            }
        }

        public void UpdateDensities()
        {
            tick++;
            bounds = new Bounds(game.Camera.Position, game.Camera.ViewWidth, game.Camera.ViewHeight);
            ICollection<Mob> mobs = game.MobsInBounds(bounds);

            List<LifeBlob> blobs = new List<LifeBlob>();
            List<Kob> kobs = new List<Kob>();
            List<Wand> wands = new List<Wand>();
            List<Crystal> crystals = new List<Crystal>();
            foreach (Mob mob in mobs)
            {
                if (mob is LifeBlob)
                    blobs.Add((LifeBlob)mob);
                if (mob is Kob)
                    kobs.Add((Kob)mob);
                if (mob is Wand)
                    wands.Add((Wand)mob);
                if (mob is Crystal)
                    crystals.Add((Crystal)mob);
            }

            maxDensity = 0;
            minDensity = 0;



            int validX = tick % Const.LifeSkip;
            int validY = (tick / Const.LifeSkip) % Const.LifeSkip;

            for (int row = 0; row < rows; row++)
            {
                for (int col = 0; col < cols; col++)
                {
                    if (row % Const.LifeSkip == validX && col % Const.LifeSkip == validY)
                    {
                        if (allGreen)
                            densities[row, col] = Const.DruidEnergy;
                        else
                            densities[row, col] = Const.baseLife;

                        float xt = (float)col / (float)(cols - 1);
                        float yt = (float)row / (float)(rows - 1);
                        float x = xt * bounds.MinX + (1 - xt) * bounds.MaxX;
                        float y = yt * bounds.MinY + (1 - yt) * bounds.MaxY;
                        Vector2 loc = new Vector2(x, y);

                        foreach (Crystal crystal in crystals)
                        {
                            if ((crystal.GobInfo.Loc - loc).Length() < crystal.GobInfo.ViewRange)
                            {
                                float e = (float)Math.Log(crystal.CrystalInfo.GivenEnergy + 1);
                                densities[row, col] += e;
                            }
                        }

                        foreach (Kob kob in kobs)
                        {
                            float e = (float)Math.Log(kob.KobInfo.Energy + 1);
                            if (kob is Druid)
                            {
                                e = Const.DruidEnergy;
                                if (kob.Mean)
                                    e *= -Const.DruidEnergy;
                            }
                            if (e <= 0)
                                continue;
                            float d = distance(loc, kob.GobInfo.Loc);
                            densities[row, col] += d * e;
                        }

                        foreach (LifeBlob blob in blobs)
                        {
                            //float e = (float)Math.Log(blob.LifeBlobInfo.Amount + 1);
                            float e = blob.LifeBlobInfo.Amount;
                            if (e <= 0)
                                continue;
                            if (blob.Source != null && blob.Source.Mean)
                                e *= -1;
                            float d = distance(loc, blob.GobInfo.Loc);
                            densities[row, col] += d * e;
                        }

                        foreach (Wand wand in wands)
                        {
                            float e = .3f;
                            //neg for bad guys
                            float d = distance(loc, wand.GobInfo.Loc);
                            densities[row, col] += d * e;
                        }
                    }

                    maxDensity = Math.Max(maxDensity, densities[row, col]);
                    minDensity = Math.Min(minDensity,densities[row, col]);
                }
            }

            /*if (maxDensity != 0)
            {
                for (int row = 0; row < rows; row++)
                {
                    for (int col = 0; col < cols; col++)
                    {
                        if (densities[row, col] > 0)
                            densities[row, col] /= maxDensity;
                        else
                            densities[row, col] /= -minDensity;
                        densities[row, col] *= 0.5f;
                        densities[row, col] += 0.5f;
                    }
                }
            }*/
            /*
            for (int row = 0; row < rows; row++)
            {
                for (int col = 0; col < cols; col++)
                {
                    if (densities[row, col] > 0)
                    {
                        densities[row, col] /= Const.DruidEnergy;
                        densities[row, col] = Math.Min(1, densities[row, col]);
                    }
                    else
                    {
                        densities[row, col] /= Const.DruidEnergy;
                        densities[row, col] = Math.Max(-1, densities[row, col]);
                    }
                    densities[row, col] *= 0.5f;
                    densities[row, col] += 0.5f;
                }
            }
             */
        }
    }
}
