﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Media;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;
using Cursed.view;
using Cursed.input;
using Cursed.gob.tile;
using Cursed.geom;
using System.Runtime.Serialization;
using Cursed.view.renders;
using Cursed.gob.mob.kob;


namespace Cursed.gob.mob
{
    public class MobInfo : GobInfo
    {
        float mass; //for collisions, perhaps accels. applying a force would consider mass.
        bool stuck;

        Vector2 vel;
        float thrust;

        float drag;
        float maxSpeedSquared;

        float theta; //rotation
        float omega; //angular velocity

        float scale;

        float lungeDuration;
        float lungeFactor;

        bool flying;
        bool aquaphobic;

        public MobInfo()
            : base()
        {
            mass = 1;
            stuck = false;
            vel = Vector2.Zero;
            thrust = 1;
            drag = 0;
            maxSpeedSquared = 100;
            theta = 0;
            omega = 0;
            scale = .9f;
            lungeDuration = 0;
            lungeFactor = 3;
            flying = false;
            aquaphobic = false;
        }

        public float Mass
        {
            get { return mass; }
            set { mass = value; }
        }
        public bool Stuck
        {
            get { return stuck; }
            set { stuck = value; }
        }
        public Vector2 Vel
        {
            get { return vel; }
            set { vel = value; }
        }
        public float Thrust
        {
            get { return thrust; }
            set { thrust = value; }
        }
        public float Drag
        {
            get { return drag;}
            set { drag = value;}
        }
        public float MaxSpeedSquared
        {
            get { return maxSpeedSquared; }
            set { maxSpeedSquared = value; }
        }
        public float Theta
        {
            get { return theta; }
            set {
                if (value - theta > MathHelper.Pi)
                    value -= 2 * MathHelper.Pi;
                if (theta - value > MathHelper.Pi)
                    value += 2 * MathHelper.Pi;
                theta = (0.9f * theta + 0.1f * value) % (float)(MathHelper.Pi * 2);
            }
        }

        public float Omega
        {
            get { return omega; }
            set { omega = value; }
        }
        public float Scale
        {
            get { return scale; }
            set { scale = value; }
        }
        public float LungeDuration
        {
            get { return lungeDuration; }
            set { lungeDuration = value; }
        }
        public float LungeFactor
        {
            get { return lungeFactor; }
            set { lungeFactor = value; }
        }
        public bool Flying
        {
            get { return flying; }
            set { flying = value; }
        }
        public bool Aquaphobic
        {
            get { return aquaphobic; }
            set { aquaphobic = value; }
        }
    }


    public class Mob : Gob, IControllable, IFocus
    {

        //direction the control is telling this mob to move at the moment
        protected Vector2 direction;
        //this time step's force vector
        protected Vector2 force;
        //this time step's drag factor
        protected float drag;

        float lungeLeft;

        float armSpacing;

        float arm1Angle;
        float arm2Angle;
        float arm1AngleDst;
        float arm2AngleDst;
        bool armsFree;

        bool armsWaving;
        float waveAngle;
        float waveTime;

        bool armsWaddling;
        float waddleDistance;
        Vector2 lastLoc;

        RenderInfo arm1Info;
        RenderInfo arm2Info;

        bool hasHands;
        string handPreset;
        Vector2 lastNonZeroDir;
        bool faceWalk;

        public Mob(CursedGame game, MobInfo info)
            : base(game, info)
        {
            lungeLeft = 0;
            force = Vector2.Zero;
            drag = 0;
            direction = Vector2.Zero;
            if (this.renderInfo != null)
            {
                this.renderInfo.Mask = info.Mask;
            }
            arm1Angle = MathHelper.PiOver4;
            arm2Angle = -MathHelper.PiOver4;
            arm1AngleDst = MathHelper.PiOver4;
            arm2AngleDst = -MathHelper.PiOver4;
            armsFree = true;
            hasHands = false;
            armsWaving = false;
            armsWaddling = false;
            waveTime = 0;
            waddleDistance = 0;
            faceWalk = true;
            armSpacing = 0.75f;
        }
        public float ArmSpacing
        {
            get { return armSpacing; }
            set { armSpacing = value; }
        }
        public bool FaceWalk
        {
            get { return faceWalk; }
            set { faceWalk = value; }
        }

        public bool HasHands
        {
            get { return hasHands; }
            set { hasHands = value; }
        }

        public string HandPreset
        {
            get { return handPreset; }
            set { handPreset = value; }
        }

        public void SpinHands(float t)
        {
            arm1AngleDst = MobInfo.Theta + (float)MathHelper.PiOver4;
            arm2AngleDst = MobInfo.Theta - (float)MathHelper.PiOver4;
            arm1AngleDst += t * 2 * (float)MathHelper.Pi;
            arm2AngleDst += t * 2 * (float)MathHelper.Pi;
            armsFree = false;
        }

        public void RelaxHands()
        {
            armsFree = true;
            armsWaving = false;
            FaceWalk = true;
        }

        public void ClapHands(float theta)
        {
            arm1AngleDst = theta;
            arm2AngleDst = theta;
            armsFree = false;
            FaceWalk = true;
        }

        public void WaveHands(float theta)
        {
            armsWaving = true;
            armsWaddling = false;
            armsFree = false;
            waveAngle = theta;
            FaceWalk = false;
        }

        public void WaddleHands()
        {
            armsWaddling = true;
            armsWaving = false;
            armsFree = false;
            FaceWalk = true;
        }

        void updateArmRender(RenderInfo armInfo, float theta)
        {
            if (armInfo == null) return;
            armInfo.Transform = Microsoft.Xna.Framework.Matrix.CreateScale(MobInfo.Scale * .25f);
            armInfo.Transform *= Microsoft.Xna.Framework.Matrix.CreateTranslation(new Vector3(0, MobInfo.Scale * armSpacing, 0));
            armInfo.Transform *= Microsoft.Xna.Framework.Matrix.CreateRotationZ(theta - (float)MathHelper.PiOver2);
            armInfo.Transform *= Microsoft.Xna.Framework.Matrix.CreateTranslation(MobInfo.Loc3);
        }


        public override void Load()
        {
            base.Load();

            if (HasHands)
            {
                arm1Info = RenderPresets.GetPreset(handPreset);
                arm1Info.Mask = renderInfo.Mask;
                addRender(arm1Info.Render, arm1Info);
                arm2Info = RenderPresets.GetPreset(handPreset);
                arm2Info.Mask = renderInfo.Mask;
                addRender(arm2Info.Render, arm2Info);
            }
        }

        public RenderInfo RenderInfo
        {
            get { return renderInfo; }
        }

        public override void Update(GameTime gameTime)
        {
            float seconds = gameTime.ElapsedGameTime.Milliseconds*.001f;
            LungeLeft -= seconds;

            if (renderInfo != null)
            {
                updateRenderTransform(renderInfo);
            }

            //System.Console.WriteLine(GetType().Name + " " + GobInfo.Loc);

            updateTransform();

            float friction = 0;

            float maxGroundDrag = 0;
            if (!MobInfo.Flying)
            {
                bool isBridged = false;

                foreach (Gob touchedGob in touchedGobs)
                {
                    if (touchedGob is Bridge)
                    {
                        isBridged = true;
                        break;
                    }
                }
                foreach (Gob touchedGob in touchedGobs)
                {
                    if (touchedGob is Ground)
                    {
                        Ground ground = (Ground)touchedGob;
                        if (ground is Water && isBridged)
                            continue;
                        maxGroundDrag = Math.Max(maxGroundDrag, ground.GroundInfo.Drag);
                        friction = Math.Max(friction, ground.GroundInfo.Friction);
                    }
                }
            }
            else
            {
                friction = 1;
            }
            
            drag += maxGroundDrag;

            float thrust = MobInfo.Thrust;
            if (LungeLeft > 0)
                thrust *= MobInfo.LungeFactor;
            ApplyForce(direction * thrust * friction);

            Vector2 dragForce = -1 * drag * MobInfo.Vel;
            drag = MobInfo.Drag;
            float maxDrag = MobInfo.Vel.Length() * MobInfo.Mass / seconds;
            if (dragForce.Length() > maxDrag)
            {
                dragForce.Normalize();
                dragForce *= maxDrag;
            }

            ApplyForce(dragForce);


            if (direction != Vector2.Zero)
                lastNonZeroDir = direction;

            if (FaceWalk)
            {
                float theta = (float)Math.Atan2(lastNonZeroDir.Y, lastNonZeroDir.X);
                if (lastNonZeroDir == Vector2.Zero)
                    theta = 0;
                MobInfo.Theta = theta;
            }

            if (HasHands)
            {
                if (armsFree)
                {
                    arm1AngleDst = MobInfo.Theta + (float)MathHelper.PiOver4;
                    arm2AngleDst = MobInfo.Theta - (float)MathHelper.PiOver4;
                }

                if (armsWaving)
                {
                    waveTime += seconds;
                    arm1AngleDst = waveAngle + (float)MathHelper.PiOver4;
                    arm2AngleDst = waveAngle - (float)MathHelper.PiOver4;
                    arm1AngleDst += .2f * (float)Math.Sin(waveTime * 20);
                    arm2AngleDst -= .2f * (float)Math.Sin(waveTime * 20);
                }

                if (armsWaddling)
                {
                    waddleDistance += (GobInfo.Loc - lastLoc).Length();
                    lastLoc = GobInfo.Loc;
                    float ot = (float)MathHelper.Pi * 3 / 8;
                    arm1AngleDst = MobInfo.Theta + ot;
                    arm2AngleDst = MobInfo.Theta - ot;
                    arm1AngleDst += .4f * (float)Math.Sin(waddleDistance * 2);
                    arm2AngleDst += .4f * (float)Math.Sin(waddleDistance * 2);
                }

                if (arm1AngleDst - arm1Angle > MathHelper.Pi)
                    arm1AngleDst -= 2 * MathHelper.Pi;
                if (arm1Angle - arm1AngleDst > MathHelper.Pi)
                    arm1AngleDst += 2 * MathHelper.Pi;
                arm1Angle = 0.7f * arm1Angle + 0.3f * arm1AngleDst;
                arm1Angle = arm1Angle % (2 * MathHelper.Pi);

                if (arm2AngleDst - arm2Angle > MathHelper.Pi)
                    arm2AngleDst -= 2 * MathHelper.Pi;
                if (arm2Angle - arm2AngleDst > MathHelper.Pi)
                    arm2AngleDst += 2 * MathHelper.Pi;
                arm2Angle = 0.7f * arm2Angle + 0.3f * arm2AngleDst;
                arm2Angle = arm2Angle % (2 * MathHelper.Pi);

                updateArmRender(arm1Info, arm1Angle);
                updateArmRender(arm2Info, arm2Angle);
            }

            base.Update(gameTime);
        }

        public MobInfo MobInfo
        {
            get { return (MobInfo)info; }
        }

        public void ApplyForce(Vector2 f)
        {
            force += f;
        }

        public Vector2 Force
        {
            get { return force; }
            set { force = value; }
        }

        public void ApplyDrag(float d)
        {
            this.drag += d;
        }

        public override bool CanTouch(Gob other)
        {
            if (other is Wall)
            {
                Wall wall = (Wall)other;
                if (MobInfo.Aquaphobic && other is Water)
                    return true;
                if (MobInfo.Flying && (other is Water || wall.WallInfo.IsShort))
                    return false;
                if (other is Water)
                {
                    if (!((Water)other).WaterInfo.Deep || ((Water)other).IsBridged)
                        return false;
                    System.Console.WriteLine("touching water");
                }
            }
            return base.CanTouch(other);
        }

        public void doPhysicsCollision(Gob gob)
        {
            Vector2 refVelocity = Vector2.Zero;
            Vector2 relVelocity = MobInfo.Vel;
            if (gob is Mob)
            {
                refVelocity = ((Mob)gob).MobInfo.Vel;
            }
            relVelocity -= refVelocity;

            Vector2 normal = Shape.intersectionNormal(gob.Shape);
            Vector2 unit = normal;
            unit.Normalize();

            float relInlineSpeed = Vector2.Dot(unit, relVelocity);

            if (relInlineSpeed < 0)
            {
                Vector2 inVel = unit * relInlineSpeed;
                Vector2 offVel = relVelocity - inVel;
                
                MobInfo.Vel = refVelocity + offVel;


                float m1 = 0;
                float m2 = 1;
                if (gob is Mob && !((Mob)gob).MobInfo.Stuck)
                {
                    Mob mob = (Mob)gob;
                    m1 = MobInfo.Mass;
                    m2 = mob.MobInfo.Mass;
                }

                MobInfo.Vel += inVel * (m1 - Const.RestitutionCoeff * m2) / (m1 + m2);

                if (gob is Mob && !((Mob)gob).MobInfo.Stuck)
                {
                    Mob mob = (Mob)gob;
                    mob.MobInfo.Vel += inVel * (Const.RestitutionCoeff + 1) * m1 / (m1 + m2);
                }
            }

            //CursedGame.Profiler.Start("pushOff");
            //pushOff(gob, normal, 1);
            //CursedGame.Profiler.Stop("pushOff");
        }

        public void Push(float factor)
        {

            foreach (Gob touchedGob in touchedGobs)
            {
                if (touches(touchedGob) && !MobInfo.Stuck && !Transcendent && !touchedGob.Transcendent)
                {
                    Vector2 normal = Shape.intersectionNormal(touchedGob.Shape);
                    pushOff(touchedGob, normal, factor);
                    updateTransform();
                    if (touchedGob is Mob)
                    {
                        ((Mob)touchedGob).updateTransform();
                    }
                }
            }
        }

        /*
         * Push self off of gob in small increments, until no longer touching
         */
        protected void pushOff(Gob gob, Vector2 normal, float factor)
        {

            if (normal == Vector2.Zero)
                return;// normal = new Vector2((float)CursedGame.Random.NextDouble(), (float)CursedGame.Random.NextDouble());

            Vector2 relV = MobInfo.Vel;
            if (gob is Mob)
                relV -= ((Mob)gob).MobInfo.Vel;
            if (Vector2.Dot(relV, normal) > 0)
                return;

            Vector2 offset = GobInfo.Loc - gob.GobInfo.Loc;
            if (Vector2.Dot(offset, normal) < 0)
                return;

            Vector2 pushDistance = normal * factor;
            if (gob is Mob && !((Mob)gob).MobInfo.Stuck)
            {
                pushDistance *= .5f;
                gob.GobInfo.Loc -= pushDistance;
            }
            GobInfo.Loc += pushDistance;
            
        }

        protected void updateTransform()
        {
            if (shape != null)
                shape.setTransform(MobInfo.Loc, MobInfo.Theta, MobInfo.Scale);
        }

        protected virtual void updateRenderTransform(RenderInfo rInfo)
        {
            rInfo.Transform = Microsoft.Xna.Framework.Matrix.CreateScale(MobInfo.Scale);
            rInfo.Transform *= Microsoft.Xna.Framework.Matrix.CreateRotationZ(MobInfo.Theta);
            rInfo.Transform *= Microsoft.Xna.Framework.Matrix.CreateTranslation(MobInfo.Loc3);
        }

        public void UpdatePosition(float seconds)
        {

            /*
            foreach (Gob touchedGob in touchedGobs)
            {
                if (!MobInfo.Stuck && !Transcendent && !touchedGob.Transcendent)
                {
                    doPhysicsCollision(touchedGob);
                }
            }
            */

            //Vector2 accelaration = force / MobInfo.Mass;

            //v = v0 + at
            //MobInfo.Vel += accelaration * seconds;

            //pos = p0 + v*t + 1/2 a*t^2;
            //MobInfo.Loc += MobInfo.Vel * seconds + accelaration * .5f * seconds * seconds;

            MobInfo.Theta += MobInfo.Omega * seconds;

            force = Vector2.Zero;
        }

        protected virtual void performLunge()
        {
            if (LungeLeft < -MobInfo.LungeDuration * 3)
            {
                LungeLeft = MobInfo.LungeDuration;
            }
        }

        public virtual void control(ControlInfo cInfo)
        {
            if (cInfo is DirectionControlInfo)
            {
                direction = ((DirectionControlInfo)cInfo).Direction;
            }
            if (cInfo is GamePadInfo && direction == Vector2.Zero)
            {
                  direction = ((GamePadInfo)cInfo).LeftStick;
            }
            if (cInfo is LungeInfo)
            {
                performLunge();
            }
        }

        public float LungeLeft
        {
            get { return lungeLeft; }
            set { lungeLeft = value; }
        }

        public bool Lunging
        {
            get { return lungeLeft > 0; }
        }
        
        #region IFocus Members

        public Vector2 Loc
        {
            get { return MobInfo.Loc; }
        }

        #endregion
    }
}
