﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Cursed.gob.tile;
using Cursed.gob;
using Cursed.gob.mob;
using Cursed.geom;

namespace Cursed.input.ai
{
    public class PathFinder
    {
        struct TCoord
        {
            public int x;
            public int y;
            public TCoord(int x, int y)
            {
                this.x = x;
                this.y = y;
            }
            public TCoord(Vector2 loc)
            {
                x = (int)Math.Floor(loc.X + .5f);
                y = (int)Math.Floor(loc.Y + .5f);
            }
            public Vector2 Vector2
            {
                get { return new Vector2(x, y); }
                set
                {
                    x = (int)Math.Floor(value.X + .5f);
                    y = (int)Math.Floor(value.Y + .5f);
                }
            }
            public bool Equals(TCoord second)
            {
                return x == second.x && y == second.y;
            }
        }

        CursedGame game;

        float epsilon;
        bool active;
        Vector2 destination;
        int depth;
        Dictionary<TCoord, float> scoreTable;
        int forgetThreshold;
        int ticks;
        int cycles;
        HashSet<Vector2> dangerPoints;

        public PathFinder(CursedGame game, int cycles)
            : this(game)
        {
            this.cycles = cycles;
        }

        public PathFinder(CursedGame game)
        {
            this.game = game;
            cycles = 1;
            epsilon = .1f;
            active = false;
            destination = Vector2.Zero;
            depth = 15;
            forgetThreshold = 100;
            ticks = 0;
            scoreTable = new Dictionary<TCoord, float>();
            dangerPoints = new HashSet<Vector2>();
        }


        public Vector2 Destination
        {
            get { return destination; }
            set { destination = value; }
        }

        TCoord DestinationTC
        {
            get { return new TCoord(destination); }
        }

        public int Depth
        {
            get { return depth; }
            set { depth = value; }
        }

        public int ForgetThreshold
        {
            get { return forgetThreshold; }
            set { forgetThreshold = value; }
        }

        protected void forget()
        {
            scoreTable.Clear();
            dangerPoints.Clear();
        }

        bool isPassable(TCoord tc)
        {
            foreach (Tile tile in game.TilesAt(tc.x, tc.y))
            {
                if (!tile.Transcendent && !(tile is Door))
                    return false;
            }
            return true;
        }

        public void AddDanger(Vector2 loc)
        {
            dangerPoints.Add(loc);
        }
        public void ClearDanger()
        {
            dangerPoints.Clear();
        }

        float getDanger(TCoord tc)
        {
            float danger = 0;
            foreach (Vector2 dangerPoint in dangerPoints)
            {
                float d2 = (tc.Vector2 - dangerPoint).LengthSquared() + 1;
                danger += 100 / d2;
            }
            return danger;
        }

        float obstruction(TCoord tc)
        {
            foreach (Tile tile in game.TilesAt(tc.x, tc.y))
            {
                if (!tile.Transcendent && !(tile is Door))
                    return Single.PositiveInfinity;
            }
            foreach (Tile tile in game.TilesAt(tc.x, tc.y-1))
            {
                if (!tile.Transcendent && !(tile is Door))
                    return Single.PositiveInfinity;
            }
            float obs = 0;
            foreach (Mob mob in game.MobsInBounds(new Bounds(tc.Vector2, 1, 1)))
            {
                if (mob.Transcendent || mob == game.DruidPlayer)
                    continue;
                if (mob.MobInfo.Stuck)
                    obs += 100;
                obs += 5f;
            }
            return obs;
        }

        float getScore(TCoord tc)
        {
            if (tc.Equals(DestinationTC))
                return 0;
            if (scoreTable.ContainsKey(tc))
                return scoreTable[tc];
            float obs = 0;
            if (!DestinationTC.Equals(tc))
                obs += obstruction(tc);
            return scoreTable[tc] = (tc.Vector2 - Destination).Length() + obs;// +getDanger(tc);
        }

        TCoord ARTDPStep(TCoord tc)
        {
            if (tc.Equals(DestinationTC))
                return tc;

            float bestScore = Single.PositiveInfinity;
            TCoord bestTC = tc;

            for (int x = tc.x - 1; x <= tc.x + 1; x++)
            {
                for (int y = tc.y - 1; y <= tc.y + 1; y++)
                {
                    TCoord n = new TCoord(x, y);
                    float ns = getScore(n);
                    if (ns < bestScore)
                    {
                        bestScore = ns;
                        bestTC = n;
                    }
                }
            }

            if (getScore(tc) != Single.PositiveInfinity)
            {
                if (bestTC.Equals(tc))
                    scoreTable[tc] = Single.PositiveInfinity;
                else
                    scoreTable[tc] = bestScore + (tc.Vector2 - bestTC.Vector2).Length();
            }

            TCoord destTC = bestTC;
            if (CursedGame.Random.NextDouble() < epsilon)
            {
                destTC.x = CursedGame.Random.Next(-1, 2);
                destTC.y = CursedGame.Random.Next(-1, 2);
            }
            if (destTC.Equals(tc))
                destTC = bestTC;

            return destTC;
        }

        protected void runTrajectory(Vector2 start)
        {
            TCoord curTC = new TCoord(start);

            int stepsLeft = Depth;
            while (stepsLeft != 0 && !curTC.Equals(DestinationTC))
            {
                curTC = ARTDPStep(curTC);
                stepsLeft--;
            }
        }

        public Vector2 GetDirection(Vector2 start)
        {
            ticks++;
            if (ticks % forgetThreshold == 0)
                forget();
            //do cycles runs of ARTDP for each act called, stopping on destination or depth threshold
            for (int i=0; i<cycles; i++)
                runTrajectory(start);

            //check what the next step for the current situation is
            TCoord startTC = new TCoord(start);
            TCoord shortTermTC = ARTDPStep(startTC);
            Vector2 diff = shortTermTC.Vector2 - start;
            if (shortTermTC.Equals(DestinationTC) || shortTermTC.Equals(startTC))
                diff = Destination - start;
            if (diff != Vector2.Zero)
                diff.Normalize();
            return diff;
        }
    }
}
