﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Cursed.gob.tile;
using Cursed.gob.mob;
using Cursed.gob.mob.kob;
using System.Collections;

namespace Cursed.input.ai.rl
{
    public class ARTDPTrajectoryHead
    {
        public State InitialState;
        public int Depth;
        public int Count;
    }

    public class MCTSDaemon : Controller
    {
        public static MCTSDaemon self = null;

        public static MCTSDaemon Self
        {
            get
            {
                if (self != null)
                    return self;
                self = new MCTSDaemon(CursedGame.TheGame);
                CursedGame.TheGame.AddController(self);
                return self;
            }
        }

        CursedGame game;

        TaskQueue<ARTDPTrajectoryHead> taskQueue;

        int minX, minY;
        float[,] rewards;
        QTable q;

        float gamma;
        float epsilon;

        HashSet<Demon> demons;

        Hashtable visits;

        public MCTSDaemon(CursedGame game)
        {
            self = this;
            demons = new HashSet<Demon>();

            this.game = game;
            q = new QTable();

            gamma = 0.9f;
            epsilon = 0.1f;

            visits = new Hashtable();

            rewards = null;

            taskQueue = new TaskQueue<ARTDPTrajectoryHead>();
            taskQueue.ProcessTask += RunTrajectoryTask;
            taskQueue.MaxTasks = 100;

            taskQueue.Start();
        }

        public void AddDemon(Demon demon)
        {
            demons.Add(demon);
        }

        public void RemoveDemon(Demon demon)
        {
            demons.Remove(demon);
        }

        public void UpdateRewards()
        {

            minX = Int32.MaxValue;
            minY = Int32.MaxValue;
            int maxX = -1 * Int32.MaxValue;
            int maxY = -1 * Int32.MaxValue;
            foreach (Tile tile in game.Tiles)
            {
                minX = Math.Min(minX, tile.TileInfo.X);
                minY = Math.Min(minY, tile.TileInfo.Y);
                maxX = Math.Max(maxX, tile.TileInfo.X);
                maxY = Math.Max(maxY, tile.TileInfo.Y);
            }
            int spanX = maxX - minX + 1;
            int spanY = maxY - minY + 1;
            rewards = new float[spanX, spanY];


            for (int x = 0; x < rewards.GetLength(0); x++)
            {
                for (int y = 0; y < rewards.GetLength(1); y++)
                {
                    rewards[x, y] = -1;
                }
            }

            foreach (Tile tile in game.Tiles)
            {
                int ox = tile.TileInfo.X - minX;
                int oy = tile.TileInfo.Y - minY;
                if (tile is Wall && !(tile is Door))
                {
                    rewards[ox, oy] = Single.NegativeInfinity;
                    if (oy+1 < rewards.GetLength(1))
                        rewards[ox, oy+1] = Single.NegativeInfinity;
                }
            }
            /*
            foreach (Mob mob in game.Mobs)
            {
                int ox = (int)Math.Floor(mob.MobInfo.Loc.X) - minX;
                int oy = (int)Math.Floor(mob.MobInfo.Loc.Y) - minY;
                if (mob.MobInfo.Stuck)
                {
                    rewards[ox, oy] -= 1f;
                }
                if (!mob.Transcendent)
                {
                    rewards[ox, oy] -= .1f;
                }
            }
             */
        }

        public DruidHunterState State(Vector2 druid, Vector2 familiar, Vector2 demon)
        {
            DruidHunterState state = new DruidHunterState(minX, minY, rewards);

            state.Druid = druid;
            state.Familiar = familiar;
            state.Demon = demon;

            return state;
        }
        public Action GetAction(Vector2 druid, Vector2 familiar, Vector2 demon)
        {
            State initialState = State(druid, familiar, demon);
            return GetAction(initialState);
        }

        public Action GetAction(State state)
        {
            return q.GetAction(state);
        }

        public void RunTrajectoryTask(ARTDPTrajectoryHead head)
        {
            for (int i=0; i<head.Count; i++)
                RunTrajectory(head.InitialState, head.Depth);
        }

        public void RunTrajectory(State current, int depth)
        {
            if (depth == 0)
                return;

            int currentVisits = 0;
            if (visits.ContainsKey(current.Index))
                currentVisits = (int)visits[current.Index];

            if (currentVisits > 200)
                return;

            visits[current.Index] = currentVisits + 1;

            Action action;
            if (game.RandomFloat < epsilon)
                action = (Action)game.RandomInt((int)Action._MAX_);
            else
                action = GetAction(current);


            float r = current.R(action);

            State next = current.Next(action);

            if (r != Single.NegativeInfinity)
                RunTrajectory(next, depth - 1);



            for (Action a = (Action)0; a < Action._MAX_; a++)
            {
                State sa = current.Next(a);
                q[current, a] = current.R(a) + gamma * q[sa];
            }

        }

        public override void act(GameTime time)
        {
            DirectionControlInfo dci = new DirectionControlInfo();

            Vector2 f = game.DruidPlayer.GobInfo.Loc;
            if (game.DruidPlayer.Familiar != null)
                f = game.DruidPlayer.Familiar.GobInfo.Loc;
            foreach (Demon demon in demons)
            {
                if (!demon.SeenMobs.Contains(game.DruidPlayer))
                {
                    dci.Direction = Vector2.Zero;
                    demon.control(dci);
                    continue;
                }

                if (rewards == null)
                    UpdateRewards();

                DruidHunterState dhs = State(game.DruidPlayer.GobInfo.Loc, f, Vector2.Zero);

                dhs.Demon = demon.GobInfo.Loc;

                ARTDPTrajectoryHead head = new ARTDPTrajectoryHead();
                head.InitialState = dhs;
                head.Depth = 15;
                head.Count = 10;

                taskQueue.PostTask(head);
                //RunTrajectoryTask(head);

                Vector2 dir = dhs.updatePosition(Vector2.Zero, GetAction(dhs));
                if (dir != Vector2.Zero)
                    dir.Normalize();
                dci.Direction = dir;
                demon.control(dci);
            }
        }

        public override void cleanup()
        {
            taskQueue.Stop();
            self = null;
        }
    }
}
